@@ -1,4 +1,6 @@
-# Copyright (C) 2018 Matthias Klumpp <matthias@tenstral.net>
+# -*- coding: utf-8 -*-
+#
+# Copyright (C) 2018-2020 Matthias Klumpp <matthias@tenstral.net>
 #
 # Licensed under the GNU Lesser General Public License Version 3
 #
@@ -23,22 +25,14 @@ from .config import GlobalConfig
 from .utils.env import set_unicode_allowed, set_owning_user
 from .osbase import OSBase
 
-__mainfile = None
-
 
 def init_config(options):
-    global __mainfile
-
+    '''
+    Create a new :GlobalConfig from command-line options.
+    '''
     gconf = GlobalConfig()
     gconf.load(options.config)
 
-    if not __mainfile.startswith('/usr'):
-        gconf.dsrun_path = os.path.normpath(os.path.join(__mainfile, '..', 'dsrun', 'dsrun.py'))
-    elif __mainfile.startswith('/usr/local'):
-        path = '/usr/local/lib/debspawn/dsrun.py'
-        if os.path.isfile(path):
-            gconf.dsrun_path = path
-
     # check if we are forbidden from using unicode - otherwise we build
     # with unicode enabled by default
     if options.no_unicode:
@@ -84,8 +78,19 @@ def command_create(options):
     if options.components:
         components = options.components.split(',')
 
-    osbase = OSBase(gconf, options.suite, options.arch, variant=options.variant, base_suite=options.base_suite)
-    r = osbase.create(options.mirror, components)
+    extra_suites = []
+    if options.extra_suites:
+        extra_suites = options.extra_suites.strip().split(' ')
+
+    osbase = OSBase(gconf,
+                    options.suite,
+                    options.arch,
+                    variant=options.variant,
+                    base_suite=options.base_suite)
+    r = osbase.create(options.mirror,
+                      components,
+                      extra_suites,
+                      options.extra_source_lines)
     if not r:
         sys.exit(2)
 
@@ -113,11 +118,24 @@ def command_update(options):
         sys.exit(1)
     gconf = init_config(options)
     osbase = OSBase(gconf, options.suite, options.arch, options.variant)
-    r = osbase.update()
+    if options.recreate:
+        r = osbase.recreate()
+    else:
+        r = osbase.update()
     if not r:
         sys.exit(2)
 
 
+def command_list(options):
+    ''' List container images '''
+
+    from .osbase import print_container_base_image_info
+
+    check_print_version(options)
+    gconf = init_config(options)
+    print_container_base_image_info(gconf)
+
+
 def command_build(options):
     ''' Build a package in a new volatile container '''
 
@@ -132,7 +150,7 @@ def command_build(options):
 
     buildflags = []
     if options.buildflags:
-        buildflags = options.buildflags.split(' ')
+        buildflags = options.buildflags.split(';')
 
     if not options.target and os.path.isdir(options.suite):
         print('A directory is given as parameter, but you are missing a suite parameter to build for.')
@@ -151,6 +169,10 @@ def command_build(options):
                                  build_only=options.build_only,
                                  include_orig=options.include_orig,
                                  maintainer=options.maintainer,
+                                 clean_source=options.clean_source,
+                                 qa_lintian=options.lintian,
+                                 interact=options.interact,
+                                 log_build=not options.no_buildlog,
                                  extra_dpkg_flags=buildflags)
     else:
         r = build_from_dsc(osbase,
@@ -159,6 +181,9 @@ def command_build(options):
                            build_only=options.build_only,
                            include_orig=options.include_orig,
                            maintainer=options.maintainer,
+                           qa_lintian=options.lintian,
+                           interact=options.interact,
+                           log_build=not options.no_buildlog,
                            extra_dpkg_flags=buildflags)
     if not r:
         sys.exit(2)
@@ -209,7 +234,6 @@ def command_run(options, custom_command):
 class CustomArgparseFormatter(HelpFormatter):
 
     def _split_lines(self, text, width):
-        print(text)
         if text.startswith('CF|'):
             return text[3:].splitlines()
         return HelpFormatter._split_lines(self, text, width)
@@ -256,6 +280,10 @@ def create_parser(formatter_class=None):
                     help='A comma-separated list of archive components to enable in the newly created image.')
     sp.add_argument('--base-suite', action='store', dest='base_suite', default=None,
                     help='A full suite that forms the base of the selected partial suite (e.g. for -updates and -backports).')
+    sp.add_argument('--extra-suites', action='store', dest='extra_suites', default=None,
+                    help='Space-separated list of additional suites that should also be added to the sources.list file.')
+    sp.add_argument('--extra-sourceslist-lines', action='store', dest='extra_source_lines', default=None,
+                    help='Lines that should be added to the build environments source.list verbatim. Separate lines by linebreaks.')
     sp.set_defaults(func=command_create)
 
     # 'delete' command
@@ -266,28 +294,43 @@ def create_parser(formatter_class=None):
     # 'update' command
     sp = subparsers.add_parser('update', help='Update a container image')
     add_container_select_arguments(sp)
+    sp.add_argument('--recreate', action='store_true', dest='recreate',
+                    help='Re-create the container image from scratch using the settings used to create it previously, instead of just updating it.')
     sp.set_defaults(func=command_update)
 
+    # 'list' command
+    sp = subparsers.add_parser('list', help='List available container images', aliases=['ls'])
+    sp.set_defaults(func=command_list)
+
     # 'build' command
-    sp = subparsers.add_parser('build', help='Build a package in an isolated environment', formatter_class=formatter_class)
+    sp = subparsers.add_parser('build', help='Build a package in an isolated environment', formatter_class=formatter_class, aliases=['b'])
     add_container_select_arguments(sp)
     sp.add_argument('-s', '--sign', action='store_true', dest='sign',
                     help='Sign the resulting package.')
     sp.add_argument('--only', choices=['binary', 'arch', 'indep', 'source'], dest='build_only',
-                    help='CF|Select only a specific set of packages to be built. Choices are:\n' +
-                         'binary: Build only binary packages, no source files are to be built and/or distributed.\n' +
-                         'arch: Build only architecture-specific binary packages.\n' +
-                         'indep: Build only architecture-independent (arch:all) binary packages.\n' +
-                         'source: Do a source-only build, no binary packages are made.')
+                    help=('CF|Select only a specific set of packages to be built. Choices are:\n'
+                          'binary: Build only binary packages, no source files are to be built and/or distributed.\n'
+                          'arch: Build only architecture-specific binary packages.\n'
+                          'indep: Build only architecture-independent (arch:all) binary packages.\n'
+                          'source: Do a source-only build, no binary packages are made.'))
     sp.add_argument('--include-orig', action='store_true', dest='include_orig',
                     help='Forces the inclusion of the original source.')
     sp.add_argument('--buildflags', action='store', dest='buildflags',
-                    help='Set flags passed through to dpkg-buildpackage.')
+                    help='Set flags passed through to dpkg-buildpackage as semicolon-separated list.')
     sp.add_argument('--results-dir', action='store', dest='results_dir',
                     help='Override the configured results directory and return artifacts at a custom location.')
     sp.add_argument('--maintainer', action='store', dest='maintainer',
-                    help='Set the name and email address of the maintainer for this package and upload, rather than using ' +
-                         'rather than using the information from the source tree\'s control file or changelog.')
+                    help=('Set the name and email address of the maintainer for this package and upload, rather than using '
+                          'the information from the source tree\'s control file or changelog.'))
+    sp.add_argument('--clean-source', action='store_true', dest='clean_source',
+                    help=('Run the d/rules clean target outside of the container. This means the package build dependencies need to be '
+                          'installed on the host system when building from a source directory.'))
+    sp.add_argument('--lintian', action='store_true', dest='lintian',
+                    help='Run the Lintian static analysis tool for Debian packages after the package is built.')
+    sp.add_argument('--no-buildlog', action='store_true', dest='no_buildlog',
+                    help='Do not write a build log.')
+    sp.add_argument('-i', '--interact', action='store_true', dest='interact',
+                    help='Run an interactive shell in the build environment after build. This implies `--no-buildlog` and disables the log.')
     sp.add_argument('target', action='store', nargs='?', default=None,
                     help='The source package file or source directory to build.')
     sp.set_defaults(func=command_build)
@@ -325,9 +368,6 @@ def run(mainfile, args):
         print('Need a subcommand to proceed!')
         sys.exit(1)
 
-    global __mainfile
-    __mainfile = mainfile
-
     parser = create_parser()
 
     # special case, so 'run' can understand which arguments are for debspawn and which are
