@@ -1,6 +1,6 @@
 # -*- coding: utf-8 -*-
 #
-# Copyright (C) 2018 Matthias Klumpp <matthias@tenstral.net>
+# Copyright (C) 2018-2020 Matthias Klumpp <matthias@tenstral.net>
 #
 # Licensed under the GNU Lesser General Public License Version 3
 #
@@ -18,11 +18,9 @@
 # along with this software.  If not, see <http://www.gnu.org/licenses/>.
 
 import os
-import subprocess
 import platform
+from .utils import temp_dir, print_error, print_warn, print_info, safe_run, run_forwarded
 from .utils.env import colored_output_allowed, unicode_allowed
-from .utils.misc import temp_dir, print_error, print_info
-from .utils.command import safe_run
 
 
 def get_nspawn_personality(osbase):
@@ -70,7 +68,7 @@ def _execute_sdnspawn(osbase, parameters, machine_name, allow_permissions=[]):
     cmd.extend(parameters)
 
     if not full_dev_access:
-        proc = subprocess.run(cmd)
+        proc = run_forwarded(cmd)
         return proc.returncode
     else:
         out, _, _ = safe_run(['systemd-escape', machine_name])
@@ -81,7 +79,7 @@ def _execute_sdnspawn(osbase, parameters, machine_name, allow_permissions=[]):
             # child process - edit the cgroup to allow full access to all
             # devices. Hopefully there won't be too much need for this awful code.
             parent_pid = os.getppid()
-            print_info('/!\ Giving container access to all host devices.')
+            print_warn('Giving container access to all host devices.')
 
             syscg_devices_allow = '/sys/fs/cgroup/devices/machine.slice/machine-{}.scope/devices.allow'.format(escaped_full_machine_name)
             tries = 0
@@ -107,11 +105,12 @@ def _execute_sdnspawn(osbase, parameters, machine_name, allow_permissions=[]):
                 sys_f.flush()
             os._exit(0)
         else:
-            proc = subprocess.run(cmd)
+            proc = run_forwarded(cmd)
             return proc.returncode
 
 
-def nspawn_run_persist(osbase, base_dir, machine_name, chdir, command=[], flags=[], tmp_apt_cache_dir=None, allowed=[], verbose=False):
+def nspawn_run_persist(osbase, base_dir, machine_name, chdir, command=[], flags=[], *,
+                       tmp_apt_cache_dir=None, pkginjector=None, allowed=[], verbose=False):
     if isinstance(command, str):
         command = command.split(' ')
     if isinstance(flags, str):
@@ -123,6 +122,9 @@ def nspawn_run_persist(osbase, base_dir, machine_name, chdir, command=[], flags=
         params = ['--chdir={}'.format(chdir),
                   '--link-journal=no',
                   '--bind={}:/var/cache/apt/archives/'.format(aptcache_tmp_dir)]
+        if pkginjector and pkginjector.instance_repo_dir:
+            params.append('--bind={}:/srv/extra-packages/'.format(pkginjector.instance_repo_dir))
+
         if personality:
             params.append('--personality={}'.format(personality))
         params.extend(flags)
@@ -172,7 +174,7 @@ def nspawn_make_helper_cmd(flags):
     if isinstance(flags, str):
         flags = flags.split(' ')
 
-    cmd = ['/usr/lib/debspawn/dsrun.py']
+    cmd = ['/usr/lib/debspawn/dsrun']
     if not colored_output_allowed():
         cmd.append('--no-color')
     if not unicode_allowed():
@@ -182,11 +184,25 @@ def nspawn_make_helper_cmd(flags):
     return cmd
 
 
-def nspawn_run_helper_ephemeral(osbase, base_dir, machine_name, helper_flags, chdir='/tmp', nspawn_flags=[], allowed=[]):
+def nspawn_run_helper_ephemeral(osbase, base_dir, machine_name, helper_flags, chdir='/tmp', *, nspawn_flags=[], allowed=[]):
     cmd = nspawn_make_helper_cmd(helper_flags)
-    return nspawn_run_ephemeral(base_dir, machine_name, chdir, cmd, nspawn_flags, allowed)
+    return nspawn_run_ephemeral(base_dir,
+                                machine_name,
+                                chdir,
+                                cmd,
+                                nspawn_flags,
+                                allowed)
 
 
-def nspawn_run_helper_persist(osbase, base_dir, machine_name, helper_flags, chdir='/tmp', nspawn_flags=[], tmp_apt_cache_dir=None, allowed=[]):
+def nspawn_run_helper_persist(osbase, base_dir, machine_name, helper_flags, chdir='/tmp', *,
+                              nspawn_flags=[], tmp_apt_cache_dir=None, pkginjector=None, allowed=[]):
     cmd = nspawn_make_helper_cmd(helper_flags)
-    return nspawn_run_persist(osbase, base_dir, machine_name, chdir, cmd, nspawn_flags, tmp_apt_cache_dir, allowed)
+    return nspawn_run_persist(osbase,
+                              base_dir,
+                              machine_name,
+                              chdir,
+                              cmd,
+                              nspawn_flags,
+                              tmp_apt_cache_dir=tmp_apt_cache_dir,
+                              pkginjector=pkginjector,
+                              allowed=allowed)
