@@ -1,6 +1,6 @@
 # -*- coding: utf-8 -*-
 #
-# Copyright (C) 2018 Matthias Klumpp <matthias@tenstral.net>
+# Copyright (C) 2018-2020 Matthias Klumpp <matthias@tenstral.net>
 #
 # Licensed under the GNU Lesser General Public License Version 3
 #
@@ -18,12 +18,14 @@
 # along with this software.  If not, see <http://www.gnu.org/licenses/>.
 
 import os
+import json
 import subprocess
 import shutil
 from pathlib import Path
 from contextlib import contextmanager
+from .utils import temp_dir, print_header, print_section, format_filesize, \
+    print_info, print_error, print_warn
 from .utils.env import ensure_root
-from .utils.misc import temp_dir, print_header, print_section, format_filesize, print_info, print_error
 from .utils.command import safe_run
 from .utils.zstd_tar import compress_directory, decompress_tarball, ensure_tar_zstd
 from .nspawn import nspawn_run_helper_persist, nspawn_run_persist
@@ -106,7 +108,7 @@ class OSBase:
     def _copy_helper_script(self, osroot_path):
         script_location = os.path.join(osroot_path, 'usr', 'lib', 'debspawn')
         Path(script_location).mkdir(parents=True, exist_ok=True)
-        script_fname = os.path.join(script_location, 'dsrun.py')
+        script_fname = os.path.join(script_location, 'dsrun')
 
         if os.path.isfile(script_fname):
             os.remove(script_fname)
@@ -117,6 +119,9 @@ class OSBase:
     def get_tarball_location(self):
         return os.path.join(self._gconf.osroots_dir, '{}.tar.zst'.format(self.name))
 
+    def get_config_location(self):
+        return os.path.join(self._gconf.osroots_dir, '{}.json'.format(self.name))
+
     def exists(self):
         return os.path.isfile(self.get_tarball_location())
 
@@ -136,26 +141,74 @@ class OSBase:
         from string import ascii_lowercase, digits
 
         nid = ''.join(choice(ascii_lowercase + digits) for _ in range(4))
-        return '{}-{}-{}'.format(platform.node(), self.name, nid)
 
-    def create(self, mirror=None, components=None):
-        ''' Create new container base image '''
-        ensure_root()
+        # on Linux, the maximum hostname length is 64, so we simple set this as general default for
+        # debspawn here.
+        # shorten the hostname part or replace the suffix, depending on what is longer.
+        # This should only ever matter if the hostname of the system already is incredibly long
+        uniq_suffix = '{}-{}'.format(self.name, nid)
+        if len(uniq_suffix) > 48:
+            uniq_suffix = ''.join(choice(ascii_lowercase + digits) for _ in range(12))
+        node_name_prefix = platform.node()[:63 - len(uniq_suffix)]
+
+        return '{}-{}'.format(node_name_prefix, uniq_suffix)
+
+    def _write_config_json(self, mirror, components, extra_suites, extra_source_lines):
+        '''
+        Create configuration file for this container base image
+        '''
+
+        print_info('Saving configuration settings.')
+        data = {'Suite': self.suite,
+                'Architecture': self.arch}
+        if self.variant:
+            data['Variant'] = self.variant
+        if mirror:
+            data['Mirror'] = mirror
+        if components:
+            data['Components'] = components
+        if extra_suites:
+            data['ExtraSuites'] = extra_suites
+        if extra_source_lines:
+            data['ExtraSourceLines'] = extra_source_lines
+
+        with open(self.get_config_location(), 'wt') as f:
+            f.write(json.dumps(data, sort_keys=True, indent=4))
+
+    def _clear_image_tree(self, image_dir):
+        ''' Clear files from a directory tree that we don't want in the tarball. '''
+
+        if os.path.ismount(image_dir):
+            print_warn('Preparing OS tree for compression, but /dev is still mounted.')
+            return
+
+        for sdir, _, files in os.walk(os.path.join(image_dir, 'dev')):
+            for f in files:
+                fname = os.path.join(sdir, f)
+                if os.path.lexists(fname) and not os.path.isdir(fname) and not os.path.ismount(fname):
+                    os.remove(fname)
+
+    def _create_internal(self, mirror=None, components=None, extra_suites=[], extra_source_lines=None, show_header=True):
+        ''' Create new container base image (internal method) '''
 
         if self.exists():
-            print_error('This configuration has already been created. You can only delete or update it.')
+            print_error('An image already exists for this configuration. Can not create a new one.')
             return False
 
         # ensure image location exists
         Path(self._gconf.osroots_dir).mkdir(parents=True, exist_ok=True)
 
-        print_header('Creating new base: {} [{}]'.format(self.suite, self.arch))
+        if show_header:
+            print_header('Creating new base: {} [{}]'.format(self.suite, self.arch))
+        else:
+            print_section('Creating new base: {} [{}]'.format(self.suite, self.arch))
+
         print('Using mirror: {}'.format(mirror if mirror else 'default'))
         if self.variant:
             print('variant: {}'.format(self.variant))
         cmd = ['debootstrap',
                '--arch={}'.format(self.arch),
-               '--include=python3']
+               '--include=python3-minimal,eatmydata']
         if components:
             cmd.append('--components={}'.format(','.join(components)))
         if self.variant:
@@ -167,6 +220,12 @@ class OSBase:
                 bootstrap_suite = self.base_suite
             cmd.extend([bootstrap_suite, tdir])
             print('Bootstrap suite: {}'.format(bootstrap_suite))
+            if extra_suites:
+                print('Additional suites: {}'.format(', '.join(extra_suites)))
+            if extra_source_lines:
+                print('Custom sources.list lines will be added:')
+                for line in extra_source_lines.split('\\n'):
+                    print('    {}'.format(line))
             if mirror:
                 cmd.append(mirror)
 
@@ -187,7 +246,7 @@ class OSBase:
                 if not mirror:
                     with open(sourceslist_fname, 'r') as f:
                         contents = f.read()
-                        matches = re.findall('http[s]?://(?:[a-zA-Z]|[0-9]|[$-_@.&+]|[!*\(\),]|(?:%[0-9a-fA-F][0-9a-fA-F]))+', contents)
+                        matches = re.findall('http[s]?://(?:[a-zA-Z]|[0-9]|[$-_@.&+]|[!*\\(\\),]|(?:%[0-9a-fA-F][0-9a-fA-F]))+', contents)
                         if not matches:
                             print_error('Unable to detect default APT repository URL (no regex matches).')
                             return False
@@ -201,16 +260,51 @@ class OSBase:
                 with open(sourceslist_fname, 'a') as f:
                     f.write('deb {mirror} {suite} {components}\n'.format(mirror=mirror, suite=self.suite, components=' '.join(components)))
 
+                    if extra_suites:
+                        f.write('\n')
+                    for esuite in extra_suites:
+                        if esuite == self.suite or esuite == bootstrap_suite:
+                            # don't add existing suites multiple times
+                            continue
+                        f.write('deb {mirror} {esuite} {components}\n'.format(mirror=mirror, esuite=esuite, components=' '.join(components)))
+
+                    if extra_source_lines:
+                        f.write('\n')
+                        for line in extra_source_lines.split('\\n'):
+                            f.write('{}\n'.format(line.strip()))
+
             print_section('Configure')
             if nspawn_run_helper_persist(self, tdir, self.new_nspawn_machine_name(), '--update') != 0:
                 return False
 
             print_section('Creating Tarball')
+            self._clear_image_tree(tdir)
             compress_directory(tdir, self.get_tarball_location())
 
-        print_info('Done.')
+        # store configuration settings, so we can later recreate this tarball
+        # or just display information about it
+        self._write_config_json(mirror, components, extra_suites, extra_source_lines)
+
         return True
 
+    def create(self, mirror=None, components=None, extra_suites=[], extra_source_lines=None):
+        ''' Create new container base image (internal method) '''
+        ensure_root()
+
+        if self.exists():
+            print_error('This configuration has already been created. You can only delete or update it.')
+            return False
+
+        ret = self._create_internal(mirror=mirror,
+                                    components=components,
+                                    extra_suites=extra_suites,
+                                    extra_source_lines=extra_source_lines,
+                                    show_header=True)
+        if ret:
+            print_info('Done.')
+
+        return ret
+
     def delete(self):
         ''' Remove container base image '''
         ensure_root()
@@ -230,6 +324,11 @@ class OSBase:
         print_section('Deleting base tarball')
         os.remove(self.get_tarball_location())
 
+        config_fname = self.get_config_location()
+        if os.path.isfile(config_fname):
+            print_section('Deleting configuration manifest')
+            os.remove(config_fname)
+
         print_info('Done.')
         return True
 
@@ -242,6 +341,9 @@ class OSBase:
     def make_instance_permanent(self, instance_dir):
         ''' Add changes done in the current instance to the main tarball of this OS tree, replacing it. '''
 
+        # remove unwanted files from the tarball
+        self._clear_image_tree(instance_dir)
+
         tarball_name = self.get_tarball_location()
         tarball_name_old = '{}.old'.format(tarball_name)
 
@@ -260,7 +362,7 @@ class OSBase:
             print_error('Can not update "{}": The configuration does not exist.'.format(self.name))
             return False
 
-        print_header('Updating container')
+        print_header('Updating container image')
 
         with self.new_instance() as (instance_dir, machine_name):
             # ensure helper script runner exists and is up to date
@@ -280,6 +382,73 @@ class OSBase:
         print_info('Done.')
         return True
 
+    def recreate(self):
+        ''' Recreate a container base image '''
+        ensure_root()
+
+        if not self.exists():
+            print_error('Can not recreate "{}": The image does not exist.'.format(self.name))
+            return False
+
+        config_fname = self.get_config_location()
+        if not os.path.isfile(config_fname):
+            print_error('Can not recreate "{}": Unable to find configuration data for this image.'.format(self.name))
+            return False
+
+        print_header('Recreating container image')
+
+        # read configuration data
+        with open(config_fname, 'rt') as f:
+            cdata = json.loads(f.read())
+            self._suite = cdata.get('Suite', self.suite)
+            self._arch = cdata.get('Architecture', self.arch)
+            self._variant = cdata.get('Variant', self.variant)
+            mirror = cdata.get('Mirror')
+            components = cdata.get('Components')
+            extra_suites = cdata.get('ExtraSuites', [])
+            extra_source_lines = cdata.get('ExtraSourceLines')
+
+        print_section('Deleting cache')
+        cache_size = self._aptcache.clear()
+        print_info('Removed {} cached packages.'.format(cache_size))
+        self._aptcache.delete()
+        print_info('Cache directory removed.')
+
+        # move old image tarball out of the way
+        image_name = self.get_tarball_location()
+        image_name_old = self.get_tarball_location() + '.old'
+        if os.path.isfile(image_name_old):
+            print_info('Removing cruft image')
+            os.remove(image_name_old)
+        os.rename(image_name, image_name_old)
+        print_info('Old tarball moved.')
+
+        # ty to create the tarball again
+        try:
+            ret = self._create_internal(mirror=mirror,
+                                        components=components,
+                                        extra_suites=extra_suites,
+                                        extra_source_lines=extra_source_lines,
+                                        show_header=False)
+        except Exception as e:
+            print_error('Error while trying to create image: {}'.format(str(e)))
+            ret = False
+
+        if ret:
+            if os.path.isfile(image_name_old):
+                print_info('Removing old image')
+                os.remove(image_name_old)
+            print_info('Done.')
+            return True
+        else:
+            print_info('Restoring old tarball')
+            if os.path.isfile(image_name):
+                print_info('Removing failed new image')
+                os.remove(image_name)
+            os.rename(image_name_old, image_name)
+            print_info('Recreation failed.')
+            return False
+
     def login(self, persistent=False, allowed=[]):
         ''' Interactive shell login into the container '''
         ensure_root()
@@ -325,6 +494,12 @@ class OSBase:
         if header_msg:
             print_header(header_msg)
 
+        # ensure we have absolute paths
+        if build_dir:
+            build_dir = os.path.abspath(build_dir)
+        if artifacts_dir:
+            artifacts_dir = os.path.abspath(artifacts_dir)
+
         with self.new_instance() as (instance_dir, machine_name):
             # ensure helper script runner exists and is up to date
             self._copy_helper_script(instance_dir)
@@ -378,3 +553,40 @@ class OSBase:
 
         print_info('Done.')
         return True
+
+
+def print_container_base_image_info(gconf):
+    '''
+    Search for all available container base images and list information
+    about them.
+    '''
+    from glob import glob
+
+    osroots_dir = gconf.osroots_dir
+    tar_files = []
+    if os.path.isdir(osroots_dir):
+        tar_files = list(glob(os.path.join(osroots_dir, '*.tar.zst')))
+    if not tar_files:
+        print_info('No container base images have been found!')
+        return False
+    tar_files_len = len(tar_files)
+
+    for i, tar_fname in enumerate(tar_files):
+        img_basepath = os.path.splitext(os.path.splitext(tar_fname)[0])[0]
+        config_fname = img_basepath + '.json'
+        imgid = os.path.basename(img_basepath)
+        print('[{}]'.format(imgid))
+
+        # read configuration data if it exists
+        if os.path.isfile(config_fname):
+            with open(config_fname, 'rt') as f:
+                cdata = json.loads(f.read())
+            for key, value in cdata.items():
+                if type(value) is list:
+                    value = '; '.join(value)
+                print('{} = {}'.format(key, value))
+
+        tar_size = os.path.getsize(tar_fname)
+        print('Size = {}'.format(format_filesize(tar_size)))
+        if i != tar_files_len - 1:
+            print()
