@@ -1,6 +1,6 @@
 # -*- coding: utf-8 -*-
 #
-# Copyright (C) 2017-2018 Matthias Klumpp <matthias@tenstral.net>
+# Copyright (C) 2017-2020 Matthias Klumpp <matthias@tenstral.net>
 #
 # Licensed under the GNU Lesser General Public License Version 3
 #
@@ -18,11 +18,10 @@
 # along with this software.  If not, see <http://www.gnu.org/licenses/>.
 
 import os
-import sys
 import shutil
 from pathlib import Path
 from contextlib import contextmanager
-from .env import unicode_allowed
+from ..config import GlobalConfig
 
 
 @contextmanager
@@ -45,7 +44,11 @@ def temp_dir(basename=None):
     else:
         dir_name = rdm_id
 
-    tmp_path = os.path.join('/var/tmp/debspawn/', dir_name)
+    temp_basedir = GlobalConfig().temp_dir
+    if not temp_basedir:
+        temp_basedir = '/var/tmp/debspawn/'
+
+    tmp_path = os.path.join(temp_basedir, dir_name)
     Path(tmp_path).mkdir(parents=True, exist_ok=True)
 
     try:
@@ -54,58 +57,6 @@ def temp_dir(basename=None):
         shutil.rmtree(tmp_path)
 
 
-def print_textbox(title, tl, hline, tr, vline, bl, br):
-    def write_utf8(s):
-        sys.stdout.buffer.write(s.encode('utf-8'))
-
-    tlen = len(title)
-    write_utf8('\n{}'.format(tl))
-    write_utf8(hline * (10 + tlen))
-    write_utf8('{}\n'.format(tr))
-
-    write_utf8('{}  {}'.format(vline, title))
-    write_utf8(' ' * 8)
-    write_utf8('{}\n'.format(vline))
-
-    write_utf8(bl)
-    write_utf8(hline * (10 + tlen))
-    write_utf8('{}\n'.format(br))
-
-    sys.stdout.flush()
-
-
-def print_header(title):
-    if unicode_allowed():
-        print_textbox(title, '╔', '═', '╗', '║', '╚', '╝')
-    else:
-        print_textbox(title, '+', '═', '+', '|', '+', '+')
-
-
-def print_section(title):
-    if unicode_allowed():
-        print_textbox(title, '┌', '─', '┐', '│', '└', '┘')
-    else:
-        print_textbox(title, '+', '-', '+', '|', '+', '+')
-
-
-def print_info(*arg):
-    '''
-    Prints an information message and ensures that it shows up on
-    stdout immediately.
-    '''
-    print(*arg)
-    sys.stdout.flush()
-
-
-def print_error(*arg):
-    '''
-    Prints an information message and ensures that it shows up on
-    stdout immediately.
-    '''
-    print('ERROR:', *arg, file=sys.stderr)
-    sys.stderr.flush()
-
-
 def format_filesize(num, suffix='B'):
     for unit in ['', 'Ki', 'Mi', 'Gi', 'Ti', 'Pi', 'Ei', 'Zi']:
         if abs(num) < 1024.0:
@@ -120,3 +71,21 @@ def current_time_string():
     from datetime import datetime, timezone
     utc_dt = datetime.now(timezone.utc)
     return utc_dt.astimezone().strftime('%Y-%m-%d %H:%M:%S UTC%z')
+
+
+def version_noepoch(version):
+    ''' Return version from :version without epoch. '''
+
+    version_noe = version
+    if ':' in version_noe:
+        version_noe = version_noe.split(':', 1)[1]
+    return version_noe
+
+
+def hardlink_or_copy(src, dst):
+    ''' Hardlink a file :src to :dst or copy the file in case linking is not possible '''
+
+    try:
+        os.link(src, dst)
+    except (PermissionError, OSError):
+        shutil.copy2(src, dst)
